@@ -1,7 +1,7 @@
 /* -*- c-basic-offset: 8 -*-
    rdesktop: A Remote Desktop Protocol client.
    Entrypoint and utility functions
-   Copyright (C) Matthew Chapman 1999-2005
+   Copyright (C) Matthew Chapman 1999-2008
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -30,10 +30,10 @@
 #include <errno.h>
 #include "rdesktop.h"
 
-#ifdef HAVE_ICONV
 #ifdef HAVE_LOCALE_H
 #include <locale.h>
 #endif
+#ifdef HAVE_ICONV
 #ifdef HAVE_LANGINFO_H
 #include <langinfo.h>
 #endif
@@ -45,13 +45,16 @@
 #include <sys/un.h>		/* sockaddr_un */
 #endif
 
-#include <openssl/md5.h>
+#include "ssl.h"
 
 char g_title[64] = "";
 char g_username[64];
 char g_hostname[16];
-char keymapname[16];
-int g_keylayout = 0x409;	/* Defaults to US keyboard layout */
+char g_keymapname[PATH_MAX] = "";
+unsigned int g_keylayout = 0x409;	/* Defaults to US keyboard layout */
+int g_keyboard_type = 0x4;	/* Defaults to US keyboard layout */
+int g_keyboard_subtype = 0x0;	/* Defaults to US keyboard layout */
+int g_keyboard_functionkeys = 0xc;	/* Defaults to US keyboard layout */
 
 int g_width = 800;		/* width is special: If 0, the
 				   geometry will be fetched from
@@ -66,31 +69,42 @@ int g_pos = 0;			/* 0 position unspecified,
 				   2 xpos neg,
 				   4 ypos neg  */
 extern int g_tcp_port_rdp;
-int g_server_bpp = 8;
+int g_server_depth = -1;
 int g_win_button_size = 0;	/* If zero, disable single app mode */
-BOOL g_bitmap_compression = True;
-BOOL g_sendmotion = True;
-BOOL g_bitmap_cache = True;
-BOOL g_bitmap_cache_persist_enable = False;
-BOOL g_bitmap_cache_precache = True;
-BOOL g_encryption = True;
-BOOL packet_encryption = True;
-BOOL g_desktop_save = True;	/* desktop save order */
-BOOL g_polygon_ellipse_orders = True;	/* polygon / ellipse orders */
-BOOL g_fullscreen = False;
-BOOL g_grab_keyboard = True;
-BOOL g_hide_decorations = False;
-BOOL g_use_rdp5 = True;
-BOOL g_console_session = False;
-BOOL g_numlock_sync = False;
-BOOL g_owncolmap = False;
-BOOL g_ownbackstore = True;	/* We can't rely on external BackingStore */
+RD_BOOL g_bitmap_compression = True;
+RD_BOOL g_sendmotion = True;
+RD_BOOL g_bitmap_cache = True;
+RD_BOOL g_bitmap_cache_persist_enable = False;
+RD_BOOL g_bitmap_cache_precache = True;
+RD_BOOL g_encryption = True;
+RD_BOOL g_packet_encryption = True;
+RD_BOOL g_desktop_save = True;	/* desktop save order */
+RD_BOOL g_polygon_ellipse_orders = True;	/* polygon / ellipse orders */
+RD_BOOL g_fullscreen = False;
+RD_BOOL g_grab_keyboard = True;
+RD_BOOL g_hide_decorations = False;
+RD_BOOL g_use_rdp5 = True;
+RD_BOOL g_rdpclip = True;
+RD_BOOL g_console_session = False;
+RD_BOOL g_numlock_sync = False;
+RD_BOOL g_lspci_enabled = False;
+RD_BOOL g_owncolmap = False;
+RD_BOOL g_ownbackstore = True;	/* We can't rely on external BackingStore */
+RD_BOOL g_seamless_rdp = False;
 uint32 g_embed_wnd;
 uint32 g_rdp5_performanceflags =
 	RDP5_NO_WALLPAPER | RDP5_NO_FULLWINDOWDRAG | RDP5_NO_MENUANIMATIONS;
+/* Session Directory redirection */
+RD_BOOL g_redirect = False;
+char g_redirect_server[64];
+char g_redirect_domain[16];
+char g_redirect_password[64];
+char g_redirect_username[64];
+char g_redirect_cookie[128];
+uint32 g_redirect_flags = 0;
 
 #ifdef WITH_RDPSND
-BOOL g_rdpsnd = False;
+RD_BOOL g_rdpsnd = False;
 #endif
 
 #ifdef HAVE_ICONV
@@ -113,7 +127,7 @@ static void
 usage(char *program)
 {
 	fprintf(stderr, "rdesktop: A Remote Desktop Protocol client.\n");
-	fprintf(stderr, "Version " VERSION ". Copyright (C) 1999-2005 Matt Chapman.\n");
+	fprintf(stderr, "Version " VERSION ". Copyright (C) 1999-2008 Matthew Chapman.\n");
 	fprintf(stderr, "See http://www.rdesktop.org/ for more information.\n\n");
 
 	fprintf(stderr, "Usage: %s [options] server[:port]\n", program);
@@ -134,6 +148,7 @@ usage(char *program)
 #ifdef HAVE_ICONV
 	fprintf(stderr, "   -L: local codepage\n");
 #endif
+	fprintf(stderr, "   -A: enable SeamlessRDP mode\n");
 	fprintf(stderr, "   -B: use BackingStore of X-server (if available)\n");
 	fprintf(stderr, "   -e: disable encryption (French TS)\n");
 	fprintf(stderr, "   -E: disable encryption from client to server\n");
@@ -164,14 +179,41 @@ usage(char *program)
 	fprintf(stderr, "         '-r printer:mydeskjet': enable printer redirection\n");
 	fprintf(stderr,
 		"             or      mydeskjet=\"HP LaserJet IIIP\" to enter server driver as well\n");
-	fprintf(stderr, "         '-r sound:[local|off|remote]': enable sound redirection\n");
+#ifdef WITH_RDPSND
+	fprintf(stderr,
+		"         '-r sound:[local[:driver[:device]]|off|remote]': enable sound redirection\n");
 	fprintf(stderr, "                     remote would leave sound on server\n");
+	fprintf(stderr, "                     available drivers for 'local':\n");
+	rdpsnd_show_help();
+#endif
+	fprintf(stderr,
+		"         '-r clipboard:[off|PRIMARYCLIPBOARD|CLIPBOARD]': enable clipboard\n");
+	fprintf(stderr, "                      redirection.\n");
+	fprintf(stderr,
+		"                      'PRIMARYCLIPBOARD' looks at both PRIMARY and CLIPBOARD\n");
+	fprintf(stderr, "                      when sending data to server.\n");
+	fprintf(stderr, "                      'CLIPBOARD' looks at only CLIPBOARD.\n");
+#ifdef WITH_SCARD
+	fprintf(stderr, "         '-r scard[:\"Scard Name\"=\"Alias Name[;Vendor Name]\"[,...]]\n");
+	fprintf(stderr, "          example: -r scard:\"eToken PRO 00 00\"=\"AKS ifdh 0\"\n");
+	fprintf(stderr,
+		"                   \"eToken PRO 00 00\" -> Device in Linux/Unix enviroment\n");
+	fprintf(stderr,
+		"                   \"AKS ifdh 0\"       -> Device shown in Windows enviroment \n");
+	fprintf(stderr, "          example: -r scard:\"eToken PRO 00 00\"=\"AKS ifdh 0;AKS\"\n");
+	fprintf(stderr,
+		"                   \"eToken PRO 00 00\" -> Device in Linux/Unix enviroment\n");
+	fprintf(stderr,
+		"                   \"AKS ifdh 0\"       -> Device shown in Windows enviroment \n");
+	fprintf(stderr,
+		"                   \"AKS\"              -> Device vendor name                 \n");
+#endif
 	fprintf(stderr, "   -0: attach to console\n");
 	fprintf(stderr, "   -4: use RDP version 4\n");
 	fprintf(stderr, "   -5: use RDP version 5 (default)\n");
 }
 
-void
+static void
 print_disconnect_reason(uint16 reason)
 {
 	char *text;
@@ -271,11 +313,17 @@ print_disconnect_reason(uint16 reason)
 	fprintf(stderr, "disconnect: %s.\n", text);
 }
 
-static BOOL
+static void
+rdesktop_reset_state(void)
+{
+	rdp_reset_state();
+}
+
+static RD_BOOL
 read_password(char *password, int size)
 {
 	struct termios tios;
-	BOOL ret = False;
+	RD_BOOL ret = False;
 	int istty = 0;
 	char *p;
 
@@ -363,20 +411,34 @@ main(int argc, char *argv[])
 	char fullhostname[64];
 	char domain[16];
 	char password[64];
-	char shell[128];
-	char directory[32];
-	BOOL prompt_password, deactivated;
+	char shell[256];
+	char directory[256];
+	RD_BOOL prompt_password, deactivated;
 	struct passwd *pw;
 	uint32 flags, ext_disc_reason = 0;
 	char *p;
 	int c;
-
+	char *locale = NULL;
 	int username_option = 0;
+	RD_BOOL geometry_option = False;
+	int run_count = 0;	/* Session Directory support */
+	RD_BOOL continue_connect = True;	/* Session Directory support */
+#ifdef WITH_RDPSND
+	char *rdpsnd_optarg = NULL;
+#endif
 
+#ifdef HAVE_LOCALE_H
+	/* Set locale according to environment */
+	locale = setlocale(LC_ALL, "");
+	if (locale)
+	{
+		locale = xstrdup(locale);
+	}
+
+#endif
 	flags = RDP_LOGON_NORMAL;
 	prompt_password = False;
 	domain[0] = password[0] = shell[0] = directory[0] = 0;
-	strcpy(keymapname, "en-us");
 	g_embed_wnd = 0;
 
 	g_num_devices = 0;
@@ -388,7 +450,7 @@ main(int argc, char *argv[])
 #endif
 
 	while ((c = getopt(argc, argv,
-			   VNCOPT "u:L:d:s:c:p:n:k:g:fbBeEmzCDKS:T:NX:a:x:Pr:045h?")) != -1)
+			   VNCOPT "Au:L:d:s:c:p:n:k:g:fbBeEmzCDKS:T:NX:a:x:Pr:045h?")) != -1)
 	{
 		switch (c)
 		{
@@ -406,6 +468,10 @@ main(int argc, char *argv[])
 				break;
 #endif
 
+			case 'A':
+				g_seamless_rdp = True;
+				break;
+
 			case 'u':
 				STRNCPY(g_username, optarg, sizeof(g_username));
 				username_option = 1;
@@ -452,10 +518,11 @@ main(int argc, char *argv[])
 				break;
 
 			case 'k':
-				STRNCPY(keymapname, optarg, sizeof(keymapname));
+				STRNCPY(g_keymapname, optarg, sizeof(g_keymapname));
 				break;
 
 			case 'g':
+				geometry_option = True;
 				g_fullscreen = False;
 				if (!strcmp(optarg, "workarea"))
 				{
@@ -515,7 +582,7 @@ main(int argc, char *argv[])
 				g_encryption = False;
 				break;
 			case 'E':
-				packet_encryption = False;
+				g_packet_encryption = False;
 				break;
 			case 'm':
 				g_sendmotion = False;
@@ -563,11 +630,13 @@ main(int argc, char *argv[])
 				break;
 
 			case 'a':
-				g_server_bpp = strtol(optarg, NULL, 10);
-				if (g_server_bpp != 8 && g_server_bpp != 16 && g_server_bpp != 15
-				    && g_server_bpp != 24)
+				g_server_depth = strtol(optarg, NULL, 10);
+				if (g_server_depth != 8 &&
+				    g_server_depth != 16 &&
+				    g_server_depth != 15 && g_server_depth != 24
+				    && g_server_depth != 32)
 				{
-					error("invalid server bpp\n");
+					error("Invalid server colour depth.\n");
 					return 1;
 				}
 				break;
@@ -578,17 +647,17 @@ main(int argc, char *argv[])
 				break;
 
 			case 'x':
-				if (strncmp("modem", optarg, 1) == 0)
+				if (str_startswith(optarg, "m"))	/* modem */
 				{
 					g_rdp5_performanceflags =
 						RDP5_NO_WALLPAPER | RDP5_NO_FULLWINDOWDRAG |
 						RDP5_NO_MENUANIMATIONS | RDP5_NO_THEMING;
 				}
-				else if (strncmp("broadband", optarg, 1) == 0)
+				else if (str_startswith(optarg, "b"))	/* broadband */
 				{
 					g_rdp5_performanceflags = RDP5_NO_WALLPAPER;
 				}
-				else if (strncmp("lan", optarg, 1) == 0)
+				else if (str_startswith(optarg, "l"))	/* lan */
 				{
 					g_rdp5_performanceflags = RDP5_DISABLE_NOTHING;
 				}
@@ -604,26 +673,31 @@ main(int argc, char *argv[])
 
 			case 'r':
 
-				if (strncmp("sound", optarg, 5) == 0)
+				if (str_startswith(optarg, "sound"))
 				{
 					optarg += 5;
 
 					if (*optarg == ':')
 					{
-						*optarg++;
+						optarg++;
 						while ((p = next_arg(optarg, ',')))
 						{
-							if (strncmp("remote", optarg, 6) == 0)
+							if (str_startswith(optarg, "remote"))
 								flags |= RDP_LOGON_LEAVE_AUDIO;
 
-							if (strncmp("local", optarg, 5) == 0)
+							if (str_startswith(optarg, "local"))
 #ifdef WITH_RDPSND
+							{
+								rdpsnd_optarg =
+									next_arg(optarg, ':');
 								g_rdpsnd = True;
+							}
+
 #else
 								warning("Not compiled with sound support\n");
 #endif
 
-							if (strncmp("off", optarg, 3) == 0)
+							if (str_startswith(optarg, "off"))
 #ifdef WITH_RDPSND
 								g_rdpsnd = False;
 #else
@@ -642,31 +716,59 @@ main(int argc, char *argv[])
 #endif
 					}
 				}
-				else if (strncmp("disk", optarg, 4) == 0)
+				else if (str_startswith(optarg, "disk"))
 				{
 					/* -r disk:h:=/mnt/floppy */
 					disk_enum_devices(&g_num_devices, optarg + 4);
 				}
-				else if (strncmp("comport", optarg, 7) == 0)
+				else if (str_startswith(optarg, "comport"))
 				{
 					serial_enum_devices(&g_num_devices, optarg + 7);
 				}
-				else if (strncmp("lptport", optarg, 7) == 0)
+				else if (str_startswith(optarg, "lspci"))
+				{
+					g_lspci_enabled = True;
+				}
+				else if (str_startswith(optarg, "lptport"))
 				{
 					parallel_enum_devices(&g_num_devices, optarg + 7);
 				}
-				else if (strncmp("printer", optarg, 7) == 0)
+				else if (str_startswith(optarg, "printer"))
 				{
 					printer_enum_devices(&g_num_devices, optarg + 7);
 				}
-				else if (strncmp("clientname", optarg, 7) == 0)
+				else if (str_startswith(optarg, "clientname"))
 				{
 					g_rdpdr_clientname = xmalloc(strlen(optarg + 11) + 1);
 					strcpy(g_rdpdr_clientname, optarg + 11);
 				}
+				else if (str_startswith(optarg, "clipboard"))
+				{
+					optarg += 9;
+
+					if (*optarg == ':')
+					{
+						optarg++;
+
+						if (str_startswith(optarg, "off"))
+							g_rdpclip = False;
+						else
+							cliprdr_set_mode(optarg);
+					}
+					else
+						g_rdpclip = True;
+				}
+				else if (strncmp("scard", optarg, 5) == 0)
+				{
+#ifdef WITH_SCARD
+					scard_enum_devices(&g_num_devices, optarg + 5);
+#else
+					warning("Not compiled with smartcard support\n");
+#endif
+				}
 				else
 				{
-					warning("Unknown -r argument\n\n\tPossible arguments are: comport, disk, lptport, printer, sound\n");
+					warning("Unknown -r argument\n\n\tPossible arguments are: comport, disk, lptport, printer, sound, clipboard, scard\n");
 				}
 				break;
 
@@ -699,6 +801,43 @@ main(int argc, char *argv[])
 	STRNCPY(server, argv[optind], sizeof(server));
 	parse_server_and_port(server);
 
+	if (g_seamless_rdp)
+	{
+		if (g_win_button_size)
+		{
+			error("You cannot use -S and -A at the same time\n");
+			return 1;
+		}
+		g_rdp5_performanceflags &= ~RDP5_NO_FULLWINDOWDRAG;
+		if (geometry_option)
+		{
+			error("You cannot use -g and -A at the same time\n");
+			return 1;
+		}
+		if (g_fullscreen)
+		{
+			error("You cannot use -f and -A at the same time\n");
+			return 1;
+		}
+		if (g_hide_decorations)
+		{
+			error("You cannot use -D and -A at the same time\n");
+			return 1;
+		}
+		if (g_embed_wnd)
+		{
+			error("You cannot use -X and -A at the same time\n");
+			return 1;
+		}
+		if (!g_use_rdp5)
+		{
+			error("You cannot use -4 and -A at the same time\n");
+			return 1;
+		}
+		g_width = -100;
+		g_grab_keyboard = False;
+	}
+
 	if (!username_option)
 	{
 		pw = getpwuid(getuid());
@@ -740,6 +879,21 @@ main(int argc, char *argv[])
 		STRNCPY(g_hostname, fullhostname, sizeof(g_hostname));
 	}
 
+	if (g_keymapname[0] == 0)
+	{
+		if (locale && xkeymap_from_locale(locale))
+		{
+			fprintf(stderr, "Autoselected keyboard map %s\n", g_keymapname);
+		}
+		else
+		{
+			STRNCPY(g_keymapname, "en-us", sizeof(g_keymapname));
+		}
+	}
+	if (locale)
+		xfree(locale);
+
+
 	if (prompt_password && read_password(password, sizeof(password)))
 		flags |= RDP_LOGON_AUTO;
 
@@ -759,30 +913,72 @@ main(int argc, char *argv[])
 
 #ifdef WITH_RDPSND
 	if (g_rdpsnd)
-		rdpsnd_init();
+	{
+		if (!rdpsnd_init(rdpsnd_optarg))
+		{
+			warning("Initializing sound-support failed!\n");
+		}
+	}
 #endif
+
+	if (g_lspci_enabled)
+		lspci_init();
+
 	rdpdr_init();
 
-	if (!rdp_connect(server, flags, domain, password, shell, directory))
-		return 1;
-
-	/* By setting encryption to False here, we have an encrypted login 
-	   packet but unencrypted transfer of other packets */
-	if (!packet_encryption)
-		g_encryption = False;
-
-
-	DEBUG(("Connection successful.\n"));
-	memset(password, 0, sizeof(password));
-
-	if (ui_create_window())
+	while (run_count < 2 && continue_connect)	/* add support for Session Directory; only reconnect once */
 	{
-		rdp_main_loop(&deactivated, &ext_disc_reason);
-		ui_destroy_window();
+		if (run_count == 0)
+		{
+			if (!rdp_connect(server, flags, domain, password, shell, directory))
+				return 1;
+		}
+		else if (!rdp_reconnect
+			 (server, flags, domain, password, shell, directory, g_redirect_cookie))
+			return 1;
+
+		/* By setting encryption to False here, we have an encrypted login 
+		   packet but unencrypted transfer of other packets */
+		if (!g_packet_encryption)
+			g_encryption = False;
+
+
+		DEBUG(("Connection successful.\n"));
+		memset(password, 0, sizeof(password));
+
+		if (run_count == 0)
+			if (!ui_create_window())
+				continue_connect = False;
+
+		if (continue_connect)
+			rdp_main_loop(&deactivated, &ext_disc_reason);
+
+		DEBUG(("Disconnecting...\n"));
+		rdp_disconnect();
+
+		if ((g_redirect == True) && (run_count == 0))	/* Support for Session Directory */
+		{
+			/* reset state of major globals */
+			rdesktop_reset_state();
+
+			STRNCPY(domain, g_redirect_domain, sizeof(domain));
+			STRNCPY(g_username, g_redirect_username, sizeof(g_username));
+			STRNCPY(password, g_redirect_password, sizeof(password));
+			STRNCPY(server, g_redirect_server, sizeof(server));
+			flags |= RDP_LOGON_AUTO;
+
+			g_redirect = False;
+		}
+		else
+		{
+			continue_connect = False;
+			ui_destroy_window();
+			break;
+		}
+
+		run_count++;
 	}
 
-	DEBUG(("Disconnecting...\n"));
-	rdp_disconnect();
 	cache_save_state();
 	ui_deinit();
 
@@ -815,11 +1011,11 @@ main(int argc, char *argv[])
 
 #ifdef EGD_SOCKET
 /* Read 32 random bytes from PRNGD or EGD socket (based on OpenSSL RAND_egd) */
-static BOOL
+static RD_BOOL
 generate_random_egd(uint8 * buf)
 {
 	struct sockaddr_un addr;
-	BOOL ret = False;
+	RD_BOOL ret = False;
 	int fd;
 
 	fd = socket(AF_UNIX, SOCK_STREAM, 0);
@@ -857,7 +1053,7 @@ generate_random(uint8 * random)
 {
 	struct stat st;
 	struct tms tmsbuf;
-	MD5_CTX md5;
+	SSL_MD5 md5;
 	uint32 *r;
 	int fd, n;
 
@@ -889,11 +1085,11 @@ generate_random(uint8 * random)
 	r[7] = st.st_ctime;
 
 	/* Hash both halves with MD5 to obscure possible patterns */
-	MD5_Init(&md5);
-	MD5_Update(&md5, random, 16);
-	MD5_Final(random, &md5);
-	MD5_Update(&md5, random + 16, 16);
-	MD5_Final(random + 16, &md5);
+	ssl_md5_init(&md5);
+	ssl_md5_update(&md5, random, 16);
+	ssl_md5_final(&md5, random);
+	ssl_md5_update(&md5, random + 16, 16);
+	ssl_md5_final(&md5, random + 16);
 }
 
 /* malloc; exit if out of memory */
@@ -909,18 +1105,42 @@ xmalloc(int size)
 	return mem;
 }
 
+/* Exit on NULL pointer. Use to verify result from XGetImage etc */
+void
+exit_if_null(void *ptr)
+{
+	if (ptr == NULL)
+	{
+		error("unexpected null pointer. Out of memory?\n");
+		exit(1);
+	}
+}
+
+/* strdup */
+char *
+xstrdup(const char *s)
+{
+	char *mem = strdup(s);
+	if (mem == NULL)
+	{
+		perror("strdup");
+		exit(1);
+	}
+	return mem;
+}
+
 /* realloc; exit if out of memory */
 void *
-xrealloc(void *oldmem, int size)
+xrealloc(void *oldmem, size_t size)
 {
 	void *mem;
 
-	if (size < 1)
+	if (size == 0)
 		size = 1;
 	mem = realloc(oldmem, size);
 	if (mem == NULL)
 	{
-		error("xrealloc %d\n", size);
+		error("xrealloc %ld\n", size);
 		exit(1);
 	}
 	return mem;
@@ -1044,7 +1264,7 @@ next_arg(char *src, char needle)
 			while (*(mvp + 1) != (char) 0x00)
 			{
 				*mvp = *(mvp + 1);
-				*mvp++;
+				mvp++;
 			}
 			*mvp = (char) 0x00;
 			p = nextval;
@@ -1082,6 +1302,124 @@ toupper_str(char *p)
 }
 
 
+RD_BOOL
+str_startswith(const char *s, const char *prefix)
+{
+	return (strncmp(s, prefix, strlen(prefix)) == 0);
+}
+
+
+/* Split input into lines, and call linehandler for each
+   line. Incomplete lines are saved in the rest variable, which should
+   initially point to NULL. When linehandler returns False, stop and
+   return False. Otherwise, return True.  */
+RD_BOOL
+str_handle_lines(const char *input, char **rest, str_handle_lines_t linehandler, void *data)
+{
+	char *buf, *p;
+	char *oldrest;
+	size_t inputlen;
+	size_t buflen;
+	size_t restlen = 0;
+	RD_BOOL ret = True;
+
+	/* Copy data to buffer */
+	inputlen = strlen(input);
+	if (*rest)
+		restlen = strlen(*rest);
+	buflen = restlen + inputlen + 1;
+	buf = (char *) xmalloc(buflen);
+	buf[0] = '\0';
+	if (*rest)
+		STRNCPY(buf, *rest, buflen);
+	strncat(buf, input, inputlen);
+	p = buf;
+
+	while (1)
+	{
+		char *newline = strchr(p, '\n');
+		if (newline)
+		{
+			*newline = '\0';
+			if (!linehandler(p, data))
+			{
+				p = newline + 1;
+				ret = False;
+				break;
+			}
+			p = newline + 1;
+		}
+		else
+		{
+			break;
+
+		}
+	}
+
+	/* Save in rest */
+	oldrest = *rest;
+	restlen = buf + buflen - p;
+	*rest = (char *) xmalloc(restlen);
+	STRNCPY((*rest), p, restlen);
+	xfree(oldrest);
+
+	xfree(buf);
+	return ret;
+}
+
+/* Execute the program specified by argv. For each line in
+   stdout/stderr output, call linehandler. Returns false on failure. */
+RD_BOOL
+subprocess(char *const argv[], str_handle_lines_t linehandler, void *data)
+{
+	pid_t child;
+	int fd[2];
+	int n = 1;
+	char output[256];
+	char *rest = NULL;
+
+	if (pipe(fd) < 0)
+	{
+		perror("pipe");
+		return False;
+	}
+
+	if ((child = fork()) < 0)
+	{
+		perror("fork");
+		return False;
+	}
+
+	/* Child */
+	if (child == 0)
+	{
+		/* Close read end */
+		close(fd[0]);
+
+		/* Redirect stdout and stderr to pipe */
+		dup2(fd[1], 1);
+		dup2(fd[1], 2);
+
+		/* Execute */
+		execvp(argv[0], argv);
+		perror("Error executing child");
+		_exit(128);
+	}
+
+	/* Parent. Close write end. */
+	close(fd[1]);
+	while (n > 0)
+	{
+		n = read(fd[0], output, 255);
+		output[n] = '\0';
+		str_handle_lines(output, &rest, linehandler, data);
+	}
+	xfree(rest);
+
+	return True;
+}
+
+
 /* not all clibs got ltoa */
 #define LTOA_BUFSIZE (sizeof(long) * 8 + 1)
 
@@ -1197,7 +1535,7 @@ save_licence(unsigned char *data, int length)
 }
 
 /* Create the bitmap cache directory */
-BOOL
+RD_BOOL
 rd_pstcache_mkdir(void)
 {
 	char *home;
@@ -1274,7 +1612,7 @@ rd_lseek_file(int fd, int offset)
 }
 
 /* do a write lock on a file */
-BOOL
+RD_BOOL
 rd_lock_file(int fd, int start, int len)
 {
 	struct flock lock;
